#ifndef  CELL_LIST_H
#define  CELL_LIST_H

#include "../GLOBAL.h"
#include "Cell.h"
#include "../box/Box.h"
#include "../particles/Particle.h"
#include <cstdlib>
#include "ParticleCellCache.h"

/**
* \defgroup CellList_Module Cell List objects
*/
/**
/**@{
* \brief    Classes used to implement a cell list. Only the CellList class
* is accessed by other classes outside this module.
*/

/**
 * \brief Structure to cache neigbour cell for given
 *
 * Used to access as qiuck as possible to neigbour cells for given i.e. without adding vectors and 
 * applying boundrary conditions to add result.
*/
typedef struct _T_CELL_LIST_H_NeighbourCellsIndexes
{
  double Cutoff; ///< cutoff to which that indexes will be applied.
#ifdef makefile_cell_kind_translation_invariant
  int CountPerCell; /**< Number of neighbour cells for given. if m=Cutoff^2/FCutoff^2 then 
  \li \c value=27 for m<=1
  \li \c value=81  for 1<m<=2 
  \li \c value=117 for 2<m<=3 
  \li \c value=125 for 3<m<=4 
  \li \c value=179 for 4<m<=5 
  \li \c value=251 for 5<m<=6 
  \li \c value=275 for 6<m<=8 
  \li \c value=331 for 8<m<=9 
  \li \c value=389 for 9<m<=10 
  and so on. Jumps only in points of kind m=x^2+y^2+z^2, where x,y,z are integers. 
  See also CellMultiplier field of CellList structure. */
#else
  int *CountsPerCells; /**< In case of non-translational lattice structures (if exists) there numbers of neighbour
                            cells for every cell stored*/
  int **CellBaseIndexes; /**< points to first neigbour for each cell*/
#endif
  int *Base; ///< assistant field, memory needed for all neighbours for all cells in one instance of T_CELL_LIST_H_NeighbourCellsIndexes structure
} T_CELL_LIST_H_NeighbourCellsIndexes;

/**
 * \brief Callback function type to proceed neighbour particles.
 *
 * There are two ways to obtain particles from cells: via copying into additional memory or via that callback.
 * Copying takes more memory used and more instructions (for at least copying itself), while callbacks have disadvantage
 * of calling function for each particle i.e. disadvantage in call-ret + prologue-epilogue additional asm instructions.
*/
typedef int T_CELL_LIST_H_ProceedCode(int PartId);

/**
* \brief A CellList maps a simulation box onto a grid of Cell objects.
*
* A CellList may be used in the calculation of nonbonded energies and
* forces to efficiently identify all particles that are within some
* cutoff distance of any point in the simulation box. 
* Time needed to modify particle in cell list is O(1) while time to obtain all elements
* from neighbour cells is O(n), where n - number of neighbors. For n=27 ineffective volume overproceeding
* = 27/(4/3*pi)~6.44
* In a CellList, a simulation box is divided into a grid of cells, such that the
* length of each cell in each of 3 directions is greater than a
* specified cutoff distance. Each cell is assigned an integer index
* between 0 and totCells - 1.
*
* \ingroup CellList_Module Particle_Module
*/
class CellList
{
public:
  /**
  * Maximum possible number of neighboring particles.
  *
  * Use as dimension of the array neighbor passed to getNeighbors() and
  * getCellNeighbors().
  */
  static const unsigned int MaxNeighborCellsIndexesCount=147; ///< maximal cutoff distance square for non-base index is =10*FCutoff^2
  static const unsigned int MaxNeighbor = MaxNeighborCellsIndexesCount *Cell::MaxPartCell; ///< maximal neighbour particles for given one
  Cell *Pcell;      ///<  array of Cell objects to store particle Ids
  TParticleCellCache *ParticleBase; ///< List of particles in this cell
  T_CELL_LIST_H_NeighbourCellsIndexes *NeighborCellsIndexes; /**< Neigbour structures for given CellList object.
  First element of structure is "primary" one. It cutoff coincides with width of cell. Other ones are secondary -
  just for quick access for neigbours in some monte carlo moves e.g. for bridge length in CbDoubleBridgeBase descedants*/
  int NeighborCellsIndexesCount; ///< Number or neighbour structures needed for current model
  /**
  * CellMultiplier is integer of kind a^2+b^2+c^2. Number of neigbour cells depends on this number.
  * Only in this kind of points there is a jumb of function of number of neigbour cells.
  **/
  static const int CellMultiplier[3]={1,1,1};
  /**
  * \brief Constructor.
  */
  CellList();
  /**
  * \brief Destructor.
  *
  * Deallocates array of Cell objects, if necessary.
  */
  ~CellList();
  /**
  * \brief Initialize the CellList grid geometry and dynamically allocate memory.
  *
  * This function:
  *   - Sets the number of cells in each dimension.
  *   - Allocates an array of Cell objects to store particle Ids.
  *   - Allocates an array of maxNParticle TParticleCellCache objects.
  *
  * The number of cells in each direction is chosen such that the dimension
  * of each cell in each direction is greater than or equal to the cutoff
  * parameter.  To calculate nonbonded pair interaction energies, the cutoff
  * parameter should thus be equal to or greater than the maximum range of
  * nonbonded interactions.
  *
  * This function does not populate the CellList with particles. See the
  * CellList::build() and CellList::addParticle() methods.
  *
  * \param Abox              pointer to Box object for the simulation box
  * \param cutoff           Minimum dimension of a cell in any direction
  * \param maxNParticle_in  maximum allowed particle Id + 1
  * \param MaxVolumeDelta        maximum volume that this Celllist can deal with (Volume might be changable in main model) 
  */
  void initialize(const Box *Abox, double cutoff, int maxNParticle_in, double MaxVolumeDelta=0.0);
  /**
  * \brief Initialize geometry
  *
  * Set number of cells and them widthes. Also set "primary" cutoff.
  * \param cutoff - "primary" cutoff of CellList object. Normally it is maximum of all potentials/forces of model.
  */
  void InitGeometry(double cutoff);
  /**
   * \brief Get all cutoffs present in CellList.
   * 
   * \param [out] OutCutoffs array of cutoffs
   * \return Number of cutoffs
  */
  int GetCutoffs(double *OutCutoffs);
  /**
   * \brief Get all neighbour indexes of CellList
   *
   * \param [out] OutIndexes array of indexes
   * \return Number of indexes
  */
  int GetCellsIndexes(T_CELL_LIST_H_NeighbourCellsIndexes **OutIndexes);
  /**
   * \brief Set all neighbour indexes of CellList
   *
   * \param [in] InIndexes array of indexes
   * \param [in] cnt Number of indexes
  */
  void SetCellsIndexes(T_CELL_LIST_H_NeighbourCellsIndexes *InIndexes, int cnt);
  /**
  * \brief Sets all Cell objects to empty state (no particles).
  */
  void clear();
  /**
  * \brief Return index of the cell that contains the position pos = (x,y,z).
  *
  * \param pos position vector, as an array pos = {x, y, z}
  * \return index of cell that contains position pos[3]
  */
  inline int cellIndexFromPosition(const Vec pos) const;
  /**
  * \brief Add a Particle to the appropriate cell, based on its position.
  *
  * \param particleId  index for Particle to be added.
  * \param pos         new particle position, as array pos = {x, y, z}
  */
  inline void addParticle(int particleId, const Vec pos);
  /**
  * \brief Add a Particle to the appropriate cell, based on rather its cellId (it's already known).
  *
  * Used primary as "semi-add" stage of some optimized routine.
  * \param particleId  index for Particle to be added.
  *  \param CellId [in] id of 
  */
  inline void LightAddParticle(int particleId, int CellId);
  /**
  * Delete a Particle object from its cell.
  *
  * \param particleId integer index for deleted particle
  */
  inline void deleteParticle(int particleId);
  /**
  * Delete a Particle object from its cell.
  *
  * \param particleId integer index for deleted particle
  */
  inline void LightDeleteParticle(int particleId);
  /**
  * \brief Update CellList to reflect a new particle position.
  *
  * If a new particle position pos[3] lies in a new cell, with a cell index
  * icell != particle.CellId, then delete the particle from the old Cell
  * object and add it to the new one. If the new cell is the same as the
  * old one (icell == particle.CellId), do nothing and return.
  *
  * This method updates the CellList, but does not update the coordinates
  * stored in the pos member of the associated Particle object. To update
  * both, use the System::moveParticle() method.
  *
  * \param particleId index of particle in particle array of the parent System
  * \param pos        array containing the new particle position
  */
  inline void updateParticleCell(int particleId, const Vec pos);
  /**
  * \brief A quick update of particle cell info based on other particle
  *
  * It quicker than usual CellList::updateParticleCell because no need to update many pointers in bidirectional list.
  * \param [in] particleIdCopyTo    index of particle, which should be updated
  * \param [in] particleIdCopyFrom  index of particle, from which an update pointers read for particle being update
  */
  inline void CopyParticleCellFromOtherOne(int particleIdCopyTo, int particleIdCopyFrom);
  /**
  * \brief An other optimized way to update particles in list - swapping
  *
  * \param [in] particleI 1st particle to swap
  * \param [in] particleJ 2nd particle to swap
  */
  inline void ChangeParticlesCells(int particleI, int particleJ);
  /**
  * Populate a CellList with particles.
  *
  * \param [in] particles  array of Particle objects
  * \param [in] nParticle number of particles
  */
  void build(Particle *const particles, int nParticle);
  /**
  * \brief Fill array neighbor with Ids of particles near a specified position.
  *
  * Upon return, array neighbor contains the integer Ids of all particles
  * in the cell containing the position pos and in all neighboring cells,
  * and nNeighbor container the number of such particles. The particle
  * Ids are in array elements neighbor[0],..., neighbor[nNeighbor-1].
  * Neighbours are taken from based (primary, zero-index) cutoff
  *
  * \param [in] pos        position array, pos= { x, y, z }
  * \param [out] neighbor   array containing Ids of neighbor particles. 
  * There should be enought memory in array to be able to get all particles. Celllist::MaxNeighbor should be quite enought.
  * \param [out] nNeighbor  number of particle Ids in array neighbor
  * \param [in] exclude   array containing Ids of excluded particles
  * \param [in] nExclude  number of particle Ids in excluded array
  */
  //inline void getNeighbors(const Vec pos, int *neighbor, int &nNeighbor) const;
  inline void getNeighbors(const Vec pos, int *neighbor, int &nNeighbor, const int *exclude, int nExclude) const;
  /**
  * Fill array with Ids of particles in a cell and all neighboring cells.
  *
  * Same as getNeighbors. But all neighbours are based on rather cell id than particle position.
  * With getCellNeighbors, getCellNeighborsByCutoff and getHalfCellNeighborsByBaseCutoff implements 
  * 1st way to get neighbour particles - via intermediate cached memory.
  *
  * \param [in] ic         cell index
  * \param [out] neighbor   array containing particle Ids
  * \param [out] nNeighbor  number of particle Ids in array neighbor
  * \param [in] exclude   array containing Ids of excluded particles
  * \param [in] nExclude  number of particle Ids in excluded array
  */
  inline void getCellNeighbors(int ic, int *neighbor, int &nNeighbor, const int *exclude, const int nExclude) const;
  /**
  * Fill array with Ids of particles in a cell and all (secondary or primary) neighboring cells.
  *
  * Same as getCellNeighbors. Neighbours are taken from raher arbitrary cached before cutoff than based one
  *
  * \param [in] Cutoff    Cutoff, neigbours for it should be calculated and cached before via CellList::BuildNeighborCellsIndexesForCutoff
  * \param [in] ic         cell index
  * \param [out] neighbor   array containing particle Ids
  * \param [out] nNeighbor  number of particle Ids in array neighbor
  * \param [in] exclude   array containing Ids of excluded particles
  * \param [in] nExclude  number of particle Ids in excluded array
  */
  inline void getCellNeighborsByCutoff(const double Cutoff, int ic, int *neighbor, int &nNeighbor,
                                       const int *exclude, const int nExclude) const;
  /**
  * \brief proceed only half of neigbour cells for particles from optimization.
  * Returns only half of cells. Because of sorting of indexes it's optimization for md (avoidinf doubles of calculation of forces)
  * \param [in] ic         cell index
  * \param [out] neighbor   array containing particle Ids
  * \param [out] nNeighbor  number of particle Ids in array neighbor
  * \return First index in return array for same cell. From there till last index will be particles from same cell
  **/
  inline int getHalfCellNeighborsByBaseCutoff(int ic, int *neighbor, int &nNeighbor) const;
  /**
  * \brief 2nd way to proceed neighbour particles - via callback
  *
  * As pointed out in T_CELL_LIST_H_ProceedCode description, this method of proceeding particles could be slow too
  * because of slow callinf function via poiter (so with prologue and epilogue)
  *
  * \param [in] ic        cell index
  * \param [in] Cutoff    cutoff, neigbours for it should be calculated and cached before via CellList::BuildNeighborCellsIndexesForCutoff
  * \param [in,out] ProceedCode callback. Should be called with current found particle id.
  */
  inline void ProceedNeighborsByCutoff(int ic, double Cutoff, T_CELL_LIST_H_ProceedCode ProceedCode) const;
  /**
  * \brief 2nd way to proceed neighbour particles - via callback
  *
  * As pointed out in T_CELL_LIST_H_ProceedCode description, this method of proceeding particles could be slow too
  * because of slow callinf function via poiter (so with prologue and epilogue)
  *
  * \param [in] pos       position of particle for which neighbours should be calculated
  * \param [in] Cutoff    cutoff, neigbours for it should be calculated and cached before via CellList::BuildNeighborCellsIndexesForCutoff
  * \param [in,out] ProceedCode callback. Should be called with current found particle id.
  */
  inline void ProceedNeighborsByCutoff(const Vec pos, double Cutoff, T_CELL_LIST_H_ProceedCode ProceedCode) const;
  /**
  * Return total number of cells in the system.
  *
  * \return total number of cells
  */
  inline int getTotCells() const
  {
    return totCells;
  }
  /**
  * \brief Check validity of CellList, and throw an Exception object if invalid.
  *
  * If the function returns normally, all tests were passed. Primary for debug usage.
  *
  * \param particles   array of Particle objects, indexed 0, ..., nParticle-1
  * \param nParticle  number of particles in particle and position arrays.
  */
  void check(Particle *const particles, int nParticle);
  /**
  * \brief Builds for all cells them neighbours accordingly to passed cutoff.
  *
  * \param ACutoff cutoff. Should be >= base (=primary, 1st) cutoff
  */
  void BuildNeighborCellsIndexesForCutoff(double ACutoff);
  /**
  *
  *
  */
  double inline GetDelta(int i)
  {
    return (numCells[i]+1)*FCutoff/(numCells[i]*cellWidth[i]);
  }
  /**
  *
  *
  */
  int inline GetNextCnt(int i)
  {
    return (numCells[i]+1);
  }
private:
  double cellWidth[3]; ///<  dimensions of each Cell in CellList grid
  int numCells[3];     ///<  Number of cells in each direction
  int YZCells;         ///<  numCells[1]*numCells[2]
  int totCells;        ///<  total number of cells in CellList
  int maxNParticle;    ///<  maximum # of particles, dimension of TParticleCellCache
  double FCutoff;      ///< primary cutoff of Celllist. 
  const Box *box;      ///< Box object that implements boundrary conditions for given Celllist structure
  /**
  * \brief Returns the number of cells required along a single Cartesian axis.
  *
  * The number of cells is chosen so that the length of each Pcell is
  * greater than or equal to the parameter cutoff.
  *
  * \param [in] axis   index of a Cartesian direction. Values: 0, 1, or 2
  */
  void setCellsAxis(int axis);
  /**
  * \brief Return integer index of a Pcell with integer coordinates (cx,cy,cz).
  *
  * \param cx  coordinate of Pcell along x direction.
  * \param cy  coordinate of Pcell along y direction.
  * \param cz  coordinate of Pcell along z direction.
  */
  inline int cellIndexFromCoord(int cx, int cy, int cz) const;
  /**
  * \brief Return (by reference) int coordinates (cx,cy,cz) of Pcell with index i.
  *
  * \param [in] i  valid cell index
  * \param [out] cx  coordinate of Pcell along x direction.
  * \param [out] cy  coordinate of Pcell along y direction.
  * \param [out] cz  coordinate of Pcell along z direction.
  */
  inline void cellCoordFromIndex(int i, int &cx, int &cy, int &cz) const;
  /**
  */
  inline double MinCellCoordDistance(double deltaR, int axis);
  void CheckOneParticle(const int partId) const;
  static inline int T_CELL_LIST_H_NeighborCellsIndexes_cmp(const void *p1, const void *p2)
  {
    double Delta=((T_CELL_LIST_H_NeighbourCellsIndexes *)p1)->Cutoff-((T_CELL_LIST_H_NeighbourCellsIndexes *)p2)->Cutoff;
    if (Delta>0.0)
      return 1;
    else if (Delta==0.0)
      return 0;
    else
      return -1;
  }
  inline const T_CELL_LIST_H_NeighbourCellsIndexes *T_CELL_LIST_H_NeighborCellsIndexes_search_by_cutoff(double ACutoff) const
  {
    /* Cannot be less than FCutoff otherwise list will be inefficient to build and... index for FCutoff
    should be first in NeighborCellsIndexes array always!
    */
    if (FCutoff>ACutoff)
      ACutoff=FCutoff;
    if (NeighborCellsIndexesCount>10)
    {
      static T_CELL_LIST_H_NeighbourCellsIndexes key;
      //
      key.Cutoff=ACutoff;
      return (T_CELL_LIST_H_NeighbourCellsIndexes *) bsearch(&key, NeighborCellsIndexes, NeighborCellsIndexesCount,
             sizeof(T_CELL_LIST_H_NeighbourCellsIndexes), T_CELL_LIST_H_NeighborCellsIndexes_cmp);
    }
    else
    {
      const T_CELL_LIST_H_NeighbourCellsIndexes *p;
      //
      for (p=NeighborCellsIndexes; p<NeighborCellsIndexes+NeighborCellsIndexesCount; p++)
        if (p->Cutoff==ACutoff)
          return p;
      return NULL;
    }
  }
  int FindParticle(const int partId) const;
  friend inline int CELL_LIST_CPP_local_int_cmp_by_coords(const void *p1, const void *p2);
  // Grant access to unit test class
  //friend class CellListTest;
};

// Public inline method definitions:
/* -----------------------------------------------------------------------------
*  Return index of the Pcell that contains the position array pos = {x, y, z}.
*/
inline int CellList::cellIndexFromPosition(const Vec pos) const
{
  return int(pos[2]/cellWidth[2]) + int(pos[1]/cellWidth[1])*numCells[2] + int(pos[0]/cellWidth[0])*YZCells;
}


/* -----------------------------------------------------------------------------
* Remove a Particle object from its Cell.
*/
inline void CellList::deleteParticle(int particleId)
{
  register TParticleCellCache *ParticleCellCache=ParticleBase+particleId;
  //
  ParticleCellCache->CellId=PMC_NULL_INDEX;
  {
    register TParticleCellCache *Next=ParticleCellCache->Next, *Prev=ParticleCellCache->Prev;
    //
    Prev->Next=Next;
    if (Next!=NULL)
      Next->Prev=Prev;
  }
}

// Reset cellTag CellId to negative "PMC_NULL_INDEX" values
// to indicate that the particle is no longer associated with a Cell.
inline void CellList::LightDeleteParticle(int particleId)
{
  (ParticleBase+particleId)->CellId=PMC_NULL_INDEX;
}

/* ----------------------------------------------------------------------------
* Add a Particle to the appropriate Pcell, based on its position. Based on assumtion that particle not present in celllist.
*/
inline void CellList::LightAddParticle(int particleId, int CellId)
{
  register TParticleCellCache *ParticleCellCache=ParticleBase+particleId, *Base=&(Pcell[CellId].Base), *Next=Base->Next;
  //
  ParticleCellCache->CellId=CellId;
  ParticleCellCache->Next=Next;
  if (Next!=NULL)
    Next->Prev=ParticleCellCache;
  ParticleCellCache->Prev=Base;
  Base->Next=ParticleCellCache;
}

inline void CellList::addParticle(int particleId, const Vec pos)
{
  int CellId = cellIndexFromPosition(pos);
  //
  LightAddParticle(particleId, CellId);
}
/* ----------------------------------------------------------------------------
* Update CellList to reflect new particle position
*/
inline void CellList::updateParticleCell(int particleId, const Vec pos)
{
  TParticleCellCache *ParticleCellCache=ParticleBase+particleId;
  int oldCell, newCell;
  //
  oldCell = ParticleCellCache->CellId;
  newCell = cellIndexFromPosition(pos);
  if (oldCell != newCell)
  {
    if (oldCell!=PMC_NULL_INDEX)
      deleteParticle(particleId);
    LightAddParticle(particleId, newCell);
  }
}
/*
 Under assumption that particleIdCopyTo is already removed from system.
 Be aware to apply this function in loop within chain - just look after direction of loop
 particleIdCopyFrom become invalid de-facto
*/
inline void CellList::CopyParticleCellFromOtherOne(int particleIdCopyTo, int particleIdCopyFrom)
{
  TParticleCellCache *ParticleCellCacheTo=ParticleBase+particleIdCopyTo, *ParticleCellCacheFrom=ParticleBase+particleIdCopyFrom,
                      *Prev=ParticleCellCacheFrom->Prev, *Next=ParticleCellCacheFrom->Next;
  //
  ParticleCellCacheTo->CellId=ParticleCellCacheFrom->CellId;
  ParticleCellCacheTo->Prev=Prev;
  Prev->Next=ParticleCellCacheTo;
  ParticleCellCacheTo->Next=Next;
  if (Next!=NULL)
    Next->Prev=ParticleCellCacheTo;
}

inline void CellList::ChangeParticlesCells(int particleI, int particleJ)
{
  TParticleCellCache *CacheI=ParticleBase+particleI, *CacheJ=ParticleBase+particleJ;
  //
  if (CacheI->CellId!=CacheJ->CellId)
  {
    int Temp;
    TParticleCellCache *Temp2;
    // change CellId:
    Temp=CacheI->CellId;
    CacheI->CellId=CacheJ->CellId;
    CacheJ->CellId=Temp;
    // change Prev:
    Temp2=CacheI->Prev;
    Temp2->Next=CacheJ;
    CacheI->Prev=CacheJ->Prev;
    CacheJ->Prev=Temp2;
    CacheI->Prev->Next=CacheI;
    // change Next:
    Temp2=CacheI->Next;
    if (Temp2!=NULL)
      Temp2->Prev=CacheJ;
    CacheI->Next=CacheJ->Next;
    CacheJ->Next=Temp2;
    if ((Temp2=CacheI->Next)!=NULL)
      Temp2->Prev=CacheI;
  }
}

/* -----------------------------------------------------------------------------
* Fill an array with Ids of particles that are near a specified position.
*/
inline void CellList::getNeighbors(const Vec pos, int *neighbor, int &nNeighbor, const int *exclude, int nExclude) const
{
  getCellNeighbors(cellIndexFromPosition(pos), neighbor, nNeighbor, exclude, nExclude);
}

/* ----------------------------------------------------------------------------
* Return integer index of a Pcell with integer coordinates (cx,cy,cz).
*/
inline int CellList::cellIndexFromCoord(int cx, int cy, int cz) const
{
  return cx*YZCells + cy*numCells[2] + cz ;
}

/* ----------------------------------------------------------------------------
* Return (by reference) int coordinates (cx,cy,cz) of Pcell with index i.
*/
inline void CellList::cellCoordFromIndex(int i, int &cx, int &cy, int &cz) const
{
  cx = (int)(i / YZCells);
  i -= cx*YZCells;
  cy = (int)(i / numCells[2]);
  cz = i - cy*numCells[2];
}

/* -----------------------------------------------------------------------------
* Fill an array with Ids of particles in Pcell ic and all neighboring cells.
*/
inline void CellList::getCellNeighborsByCutoff(const double Cutoff, int ic, int *neighbor, int &nNeighbor,
    const int *exclude, const int nExclude) const
{
  int *PNeighb, *PNeighbBase, *PNeighbEnd;
  PMC_register_var_alloc int partId, *neighbor_p;
  const int *excl;
  const T_CELL_LIST_H_NeighbourCellsIndexes *CurrIndex;
  PMC_register_var_alloc TParticleCellCache *El;
  //
  if ((CurrIndex=T_CELL_LIST_H_NeighborCellsIndexes_search_by_cutoff(Cutoff))==NULL)
    PMC_THROW("There is no index for this cutoff");
  neighbor_p=neighbor;
#ifdef makefile_cell_kind_translation_invariant
  PNeighbBase=CurrIndex->Base+ (ic*CurrIndex->CountPerCell);
  PNeighbEnd=PNeighbBase+CurrIndex->CountPerCell;
#else
  PNeighbBase=CurrIndex->CellBaseIndexes[ic];
  PNeighbEnd=PNeighbBase+CurrIndex->CountsPerCells[ic];
#endif
  for (PNeighb=PNeighbBase; PNeighb<PNeighbEnd; PNeighb++)
  {
    for (El=Pcell[*PNeighb].Base.Next; El!=NULL; El=El->Next)
    {
      partId=El->Data;
      for (excl=exclude; excl<exclude+nExclude; excl++)
        if (*excl==partId)
          goto cont;
    #ifdef debug
      CheckOneParticle(partId);
    #endif
      PMC_ASSERT(((unsigned int)(neighbor_p-neighbor))>=MaxNeighbor, "nNeighbor>=MaxNeighbor");
      *neighbor_p = partId;
      neighbor_p++;
    cont:
      ;
    } // end for jp
  }
  nNeighbor=neighbor_p-neighbor;
}

/* -----------------------------------------------------------------------------
* Fill an array with Ids of particles in Pcell ic and all neighboring cells.
*/
inline int CellList::getHalfCellNeighborsByBaseCutoff(int ic, int *neighbor, int &nNeighbor) const
{
  int  *PNeighb, *PNeighbBase, *PNeighbEnd, res=0;
  PMC_register_var_alloc int partId, *neighbor_p;
  PMC_register_var_alloc TParticleCellCache *El;
  bool DoBreak=false;
  //
  neighbor_p=neighbor;
#ifdef makefile_cell_kind_translation_invariant
  PNeighbBase=NeighborCellsIndexes->Base+ (ic*NeighborCellsIndexes->CountPerCell);
  PNeighbEnd=PNeighbBase+NeighborCellsIndexes->CountPerCell;
#else
  PNeighbBase=NeighborCellsIndexes->CellBaseIndexes[ic];
  PNeighbEnd=PNeighbBase+NeighborCellsIndexes->CountsPerCells[ic];
#endif
  for (PNeighb=PNeighbBase; PNeighb<PNeighbEnd; PNeighb++)
  {
    if (*PNeighb==ic)
    {
      res=neighbor_p-neighbor;
      DoBreak=true;
    }
    for (El=Pcell[*PNeighb].Base.Next; El!=NULL; El=El->Next)
    {
      partId=El->Data;
    #ifdef debug
      CheckOneParticle(partId);
    #endif
      PMC_ASSERT(((unsigned int)(neighbor_p-neighbor))>=MaxNeighbor, "nNeighbor>=MaxNeighbor");
      *neighbor_p = partId;
      neighbor_p++;
    } // end for jp
    if (DoBreak)
      break;
  }
  nNeighbor=neighbor_p-neighbor;
  return res;
}

inline void CellList::getCellNeighbors(int ic, int *neighbor, int &nNeighbor, const int *exclude, int nExclude) const
{
  getCellNeighborsByCutoff(NeighborCellsIndexes->Cutoff, ic, neighbor, nNeighbor, exclude, nExclude);
}

inline void CellList::ProceedNeighborsByCutoff(int ic, double Cutoff, T_CELL_LIST_H_ProceedCode ProceedCode) const
{
  const T_CELL_LIST_H_NeighbourCellsIndexes *CurrIndex;
  int *PNeighb, *PNeighbBase, *PNeighbEnd;
  PMC_register_var_alloc TParticleCellCache *El;
  //
  if ((CurrIndex=T_CELL_LIST_H_NeighborCellsIndexes_search_by_cutoff(Cutoff))==NULL)
    PMC_THROW("There is no index for this cutoff");
#ifdef makefile_cell_kind_translation_invariant
  PNeighbBase=CurrIndex->Base+ (ic*CurrIndex->CountPerCell);
  PNeighbEnd=PNeighbBase+CurrIndex->CountPerCell;
#else
  PNeighbBase=CurrIndex->CellBaseIndexes[ic];
  PNeighbEnd=PNeighbBase+CurrIndex->CountsPerCells[ic];
#endif
  for (PNeighb=PNeighbBase; PNeighb<PNeighbEnd; PNeighb++)
    for (El=Pcell[*PNeighb].Base.Next; El!=NULL; El=El->Next)
      if (ProceedCode(El->Data))
        return;
}

/* -----------------------------------------------------------------------------
* Fill an array with Ids of particles that are near a specified position.
*/
inline void CellList::ProceedNeighborsByCutoff(const Vec pos, double Cutoff, T_CELL_LIST_H_ProceedCode ProceedCode) const
{
  ProceedNeighborsByCutoff(cellIndexFromPosition(pos), Cutoff, ProceedCode);
}

/**@}*/
#endif
